"""Unit tests for germinate.version."""

# Copyright (C) 2019 Canonical Ltd.
#
# Germinate is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Germinate is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Germinate; see the file COPYING.  If not, see
# <https://www.gnu.org/licenses/>.

from germinate.tests.helpers import TestCase
from germinate.version import VERSION


class TestVersion(TestCase):
    """Ensure that version.txt was properly included in the package."""

    def test_not_local(self):
        self.assertNotEqual("local", VERSION)
